from hrepr import H

from .constructors import BrowserEvent, FormData
from .page import Component, JavaScriptOperation, Page, selector_for
from .ref import Reference
from .serve import (
    ConfigurableBear,
    ConfigurableSimpleBear,
    bear,
    dev_injections,
    simplebear,
)
from .templating import template
from .utils import ClientWrap, FeedbackQueue, Queue, VirtualFile



import weakref
from collections import deque
from itertools import count
from types import MethodType

_c = count()


def _id(id=None):
    return next(_c) if id is None else id


class StrongRegistry:
    def __init__(self):
        self.map = {}

    def register(self, obj, id=None):
        currid = _id(id)
        self.map[currid] = obj
        return currid

    def resolve(self, id):
        return self.map[id]


class StrongRotatingRegistry:
    def __init__(self, keep, rotate):
        self.keep = keep
        self.rotate = rotate
        self.map = {}
        self.ids = deque()

    def register(self, obj, id=None):
        currid = _id(id)
        self.map[currid] = obj
        self.ids.append(currid)
        if len(self.ids) > self.keep >= 0:
            if self.rotate:
                rm = self.ids.popleft()
                del self.map[rm]
            else:
                raise Exception("Exceeded limit for keeping strong references to objects.")
        return currid

    def resolve(self, id):
        return self.map[id]


class WeakRegistry:
    def __init__(self):
        self.map = {}

    def register(self, obj, id=None):
        currid = _id(id)
        if isinstance(obj, MethodType):
            ref = weakref.WeakMethod(obj)
        else:
            ref = weakref.ref(obj)
        self.map[currid] = ref
        return currid

    def resolve(self, id):
        value = self.map[id]()
        if value is None:
            raise KeyError(id)
        else:
            return value


class ObjectRegistry:
    def __init__(self, strongrefs=False, rotate_strongrefs="limit"):
        self.wr = WeakRegistry()
        if strongrefs:
            self.sr = StrongRotatingRegistry(keep=strongrefs, rotate=rotate_strongrefs)
        else:
            self.sr = None

    def register(self, obj, id=None):
        try:
            return self.wr.register(obj, id)
        except TypeError:
            return self.sr.register(obj, id)

    def resolve(self, id):
        try:
            return self.wr.resolve(id)
        except KeyError:
            return self.sr.resolve(id)


class Reference:
    def __init__(self, datum, id=None):
        self.id = _id(id)
        self.datum = datum



import weakref
from collections import deque
from hashlib import md5
from itertools import count
from pathlib import Path
from types import FunctionType, MethodType
from typing import Union

from hrepr import embed, hrepr, standard_html
from ovld import has_attribute

from .ref import ObjectRegistry, Reference, StrongRegistry
from .utils import FeedbackQueue, VirtualFile


class CallbackRegistry:
    def __init__(self, weak=False, keep=None):
        self.id = count()
        self.weak = weak
        self.keep = (keep or 10000) if self.weak else -1
        self.weak_map = {}
        self.strong_map = {}
        self.strong_ids = deque()

    def register(self, method):
        currid = next(self.id)
        weak = self.weak
        if weak:
            try:
                self.weak_map[currid] = weakref.WeakMethod(method)
            except TypeError:
                weak = False

        if not weak:
            self.strong_map[currid] = method
            self.strong_ids.append(currid)
            if len(self.strong_ids) > self.keep >= 0:
                rm = self.strong_ids.popleft()
                del self.strong_map[rm]

        return currid

    def resolve(self, id):
        try:
            m = self.weak_map[id]()
        except KeyError:
            m = self.strong_map.get(id, None)

        if m is None:
            raise KeyError(id)

        return m


class FileRegistry:
    def __init__(self):
        self.file_to_url = {}
        self.url_to_file = {}

    def find_anchor(self, filename):
        filename = filename.absolute()
        anchor = filename
        while anchor != Path("/"):
            anchor = anchor.parent
            if anchor in self.file_to_url:
                return (anchor, self.file_to_url[anchor])
        else:
            anchor = filename.parent
            while not (anchor / "starbear-anchor").exists():
                anchor = anchor.parent
                if anchor == Path("/"):
                    anchor = filename.parent
                    break
            url = md5(str(anchor).encode("utf8")).hexdigest()
            self.file_to_url[anchor] = url
            self.url_to_file[url] = anchor
            return (anchor, url)

    def register(self, filename):
        filename = filename.absolute()
        anchor, base_url = self.find_anchor(filename)
        url = str(base_url / filename.relative_to(anchor))
        return url

    def get_file_from_url(self, url):
        url = orig = Path(url)
        while url != url.parent and str(url) not in self.url_to_file:
            url = url.parent
        anchor = self.url_to_file.get(str(url), None)
        return anchor and str(anchor / orig.relative_to(url))


class VFileRegistry:
    def __init__(self):
        self.vfiles = {}

    def register(self, vfile):
        self.vfiles[vfile.name] = vfile
        return vfile.name

    def get(self, pth):
        return self.vfiles[pth]


class FutureRegistry:
    def __init__(self):
        self.current_id = count()
        self.futures = {}

    def register(self, future):
        fid = next(self.current_id)
        self.futures[fid] = future
        return fid

    def resolve(self, fid, value):
        if fid not in self.futures:
            return
        self.futures[fid].set_result(value)
        del self.futures[fid]

    def reject(self, fid, error):
        if fid not in self.futures:
            return
        self.futures[fid].set_exception(Exception(error))
        del self.futures[fid]


class QueueRegistry:
    def __init__(self):
        self.current_id = count()
        self.queues = {}

    def register(self, queue):
        qid = next(self.current_id)
        self.queues[qid] = queue
        return qid

    def put(self, qid, value):
        self.queues[qid].put_nowait(value)


class Representer:
    def __init__(self, route, strongrefs=False):
        from asyncio import Future, Queue

        representer = self

        if strongrefs:
            object_registry = self.object_registry = StrongRegistry()
        else:
            object_registry = self.object_registry = ObjectRegistry(
                strongrefs=100, rotate_strongrefs=False
            )

        file_registry = self.file_registry = FileRegistry()
        vfile_registry = self.vfile_registry = VFileRegistry()
        future_registry = self.future_registry = FutureRegistry()
        queue_registry = self.queue_registry = QueueRegistry()
        self.route = route
        self.hrepr = hrepr

        @embed.js_embed.variant
        def js_embed(self, fn: Union[MethodType, FunctionType]):
            method_id = object_registry.register(fn)
            return f"$$BEAR.func({method_id})"

        @js_embed.register
        def js_embed(self, ref: Reference):
            obj_id = object_registry.register(ref.datum)
            return f"$$BEAR.ref({obj_id})"

        @js_embed.register
        def js_embed(self, pth: Path):
            new_pth = file_registry.register(pth)
            return f"'{route}/file/{new_pth}'"

        @js_embed.register
        def js_embed(self, future: Future):
            fid = future_registry.register(future)
            return f"$$BEAR.promise({fid})"

        @js_embed.register
        def js_embed(self, queue: Queue):
            qid = queue_registry.register(queue)
            return f"$$BEAR.queue({qid})"

        @js_embed.register
        def js_embed(self, queue: FeedbackQueue):
            qid = queue_registry.register(queue)
            return f"$$BEAR.queue({qid}, true)"

        @js_embed.register
        def js_embed(self, obj: has_attribute("__js_embed__")):
            return obj.__js_embed__(representer)

        @embed.attr_embed.variant
        def attr_embed(self, attr: str, fn: Union[MethodType, FunctionType]):
            method_id = object_registry.register(fn)
            if attr.startswith("hx_"):
                return f"{route}/method/{method_id}"
            else:
                return f"$$BEAR.event.call(this, $$BEAR.func({method_id}))"

        @attr_embed.register
        def attr_embed(self, attr: str, ref: Reference):
            obj_id = object_registry.register(ref.datum, id=ref.id)
            return f"obj#{obj_id}"

        @attr_embed.register
        def attr_embed(self, attr: str, queue: Queue):
            qid = queue_registry.register(queue)
            return f"$$BEAR.event.call(this, $$BEAR.queue({qid}))"

        @attr_embed.register
        def attr_embed(self, attr: str, pth: Path):
            new_pth = file_registry.register(pth)
            return f"{route}/file/{new_pth}"

        @attr_embed.register
        def attr_embed(self, attr: str, vf: VirtualFile):
            pth = vfile_registry.register(vf)
            return f"{route}/vfile/{pth}"

        @attr_embed.register
        def attr_embed(self, attr: str, style: dict):
            if attr == "style":
                return ";".join(f"{k}:{v}" for k, v in style.items())
            else:
                raise TypeError(f"Cannot serialize a dict for attribute '{attr}'")

        @attr_embed.register
        def attr_embed(self, attr: str, obj: has_attribute("__attr_embed__")):
            return obj.__attr_embed__(representer, attr)

        self.js_embed = js_embed
        self.attr_embed = attr_embed

        self.printer = standard_html.fork(
            js_embed=js_embed,
            attr_embed=attr_embed,
        )

    def generate(self, *args, **kwargs):
        return self.printer.generate(*args, **kwargs)

    def generate_string(self, element):
        parts, extras, resources = self.printer.generate(element)
        return str(parts)

    def __call__(self, node):
        return self.hrepr(node)



import asyncio as aio
from pathlib import Path

from hrepr import H, Tag
from hrepr.hgen import ResourceDeduplicator
from hrepr.resource import Resource

from .ref import Reference
from .utils import format_error


def selector_for(x):
    if isinstance(x, Tag):
        tid = x.attributes.get("id", None)
        if not tid:
            raise Exception("Cannot locate element because it has no id.")
        return f"#{tid}"
    elif isinstance(x, str):
        return x
    elif isinstance(x, Reference):
        return f'[--ref="obj#{x.id}"]'
    elif (selector := getattr(x, "selector", None)) and isinstance(selector, str):
        return selector
    else:
        raise TypeError(
            "A valid selector must be a str, Tag, Reference or an object"
            " with a string attribute named `selector`."
        )


class Component:
    @property
    def selector(self):
        return selector_for(self.node)

    def __hrepr__(self, H, hrepr):
        return self.node


class Page:
    def __init__(
        self,
        instance,
        selector=None,
        track_history=True,
        sent_resources=None,
        debug=False,
        loop=None,
    ):
        self.instance = instance
        self.iq = instance.iq
        self.oq = instance.oq
        self.query_params = instance.query_params
        self.session = instance.session
        self.representer = instance.representer
        self.selector = selector
        self.track_history = track_history
        self.sent_resources = sent_resources or ResourceDeduplicator()
        self.tasks = set()
        self.debug = debug
        self.loop = loop or aio.get_running_loop()
        self.js = JavaScriptOperation(self, [])
        self.window = JavaScriptOperation(self, [], root="window")
        self.bearlib = JavaScriptOperation(self, [], root="$$BEAR")

    def __getitem__(self, selector):
        if not isinstance(selector, tuple):
            selector = (selector,)
        selector = " ".join(map(selector_for, selector))

        if self.selector is not None:
            selector = f"{self.selector} {selector}"

        return self.page_select(selector)

    def page_select(self, selector):
        return type(self)(
            instance=self.instance,
            selector=selector,
            track_history=self.track_history,
            sent_resources=self.sent_resources,
            debug=self.debug,
            loop=self.loop,
        )

    def with_history(self, track_history=True):
        if self.track_history == track_history:
            return self
        else:
            return type(self)(
                instance=self.instance,
                selector=self.selector,
                track_history=track_history,
                sent_resources=self.sent_resources,
                debug=self.debug,
                loop=self.loop,
            )

    def without_history(self):
        return self.with_history(False)

    def _done_cb(self, future):
        self.tasks.discard(future)
        if exc := future.exception():
            self.error(
                message="An error occurred trying to represent data.",
                exception=exc,
            )

    def _push(self, coro):
        if aio._get_running_loop() is None:
            aio._set_running_loop(self.loop)
        task = aio.create_task(coro)
        self.tasks.add(task)
        task.add_done_callback(self._done_cb)

    async def sync(self):
        for task in list(self.tasks):
            await task

    def _to_element(self, x):
        if isinstance(x, str):
            return H.span(x)
        else:
            return self.representer.hrepr(x)

    def _generate_put_commands(self, element, method, send_resources=False):
        sel = self.selector or "body"
        if not element:
            yield {
                "command": "put",
                "selector": sel,
                "method": method,
                "content": "",
            }
            return

        parts, extra, resources = self.representer.generate(
            element, filter_resources=self.sent_resources if send_resources else None
        )
        if not resources.empty():
            yield {
                "command": "resource",
                "content": str(resources),
            }
        yield {
            "command": "put",
            "selector": sel,
            "method": method,
            "content": str(parts),
        }
        if not extra.empty():
            yield {
                "command": "put",
                "selector": sel,
                "method": "beforeend",
                "content": str(H.div(extra, style="display:none")),
            }

    async def put(self, element, method, history=None, send_resources=True):
        if history is None:
            history = self.track_history
        await self.oq.put(
            (
                list(self._generate_put_commands(element, method, send_resources)),
                history,
            )
        )

    def put_nowait(self, element, method, history=None, send_resources=True):
        self._push(self.put(element, method, history=history, send_resources=send_resources))

    def queue_command(self, command, /, history=None, **arguments):
        if history is None:
            history = self.track_history
        arguments["command"] = command
        self._push(self.oq.put((arguments, history)))

    def set_title(self, title):
        self.queue_command("put", selector="head title", content=title, method="innerHTML")

    def add_resources(self, *resources, type=None):
        def _build(resource, name):
            if name.endswith(".css") or type == "text/css":
                return H.link(rel="stylesheet", href=resource)
            elif name.endswith(".js") or type == "text/javascript":
                return H.script(src=resource)
            elif name.endswith(".ico"):
                return H.link(rel="icon", href=resource)
            else:
                raise ValueError(f"Cannot determine resource type for '{resource}'")

        for resource in resources:
            if isinstance(resource, str):
                if resource.startswith("http://") or resource.startswith("https://"):
                    node = _build(resource, resource)
                else:
                    node = _build(Path(resource), resource)
            elif isinstance(resource, Path):
                node = _build(resource, resource.suffix)
            elif isinstance(resource, Tag):
                node = resource
            else:
                raise TypeError("resource argument should be a Path or a Tag object")

            parts, _, _ = self.representer.generate(node, filter_resources=None)
            self.queue_command("resource", content=str(parts))

    def print(self, *elements, method="beforeend"):
        for element in elements:
            element = self._to_element(element)
            self.put_nowait(element, method)

    def error(self, message, debug=None, exception=None):
        if not isinstance(message, str):
            message = str(self.representer.hrepr(message))
        self.queue_command("error", content=format_error(message, debug, exception, self.debug))

    def log(self, message):
        if not isinstance(message, str):
            message = self.representer.hrepr(message)
        self.queue_command("log", content=str(message))

    def print_html(self, html, selector=None, method="beforeend", history=None):
        self.queue_command(
            "put",
            selector=selector or self.selector or "body",
            method=method,
            content=html,
            history=history,
        )

    def template(self, template_file, integration_method="innerHTML", **params):
        filled = self.instance.template(template_file, **params)
        self.put_nowait(filled, integration_method)

    def set(self, element):
        element = self._to_element(element)
        self.put_nowait(element, "innerHTML")

    def replace(self, element):
        element = self._to_element(element)
        self.put_nowait(element, "outerHTML")

    def clear(self):
        self.put_nowait("", "innerHTML")

    def delete(self):
        self.put_nowait("", "outerHTML")

    def do(self, js, future=None):
        call_template = "$$BEAR.cb({selector}, {extractor}, {future});"
        orig_code = call_template.format(
            selector=Resource(self.selector),
            extractor=f"function () {{ {js} }}",
            future=Resource(future),
        )
        code = self.representer.printer.expand_resources(
            orig_code,
            self.representer.js_embed,
        )
        self.queue_command("eval", selector=self.selector, code=code, history=False)

    def toggle(self, toggle, value=None):
        return self.bearlib.toggle(self, toggle, value)

    def __hrepr__(self, H, hrepr):
        return hrepr.make.instance(
            title="Page",
            fields=[["selector", self.selector]],
            delimiter=":",
        )

    def __js_embed__(self, representer):
        if self.selector is None:
            raise Exception("Cannot send page reference to JS because it has no selector.")
        return f"document.querySelector('{self.selector}')"

    async def recv(self):
        return await self.iq.get()


def _extractor(root, sequence):
    result = root
    for entry in sequence:
        if isinstance(entry, str):
            result = f"{result}.{entry}"
        elif isinstance(entry, (list, tuple)):
            args = ",".join([str(Resource(x)) for x in entry])
            result = f"{result}({args})"
        else:
            raise TypeError()
    return f"{{ return {result}; }}"


class JavaScriptOperation:
    def __init__(self, element, sequence, root="this"):
        self.__element = element
        self.__sequence = sequence
        self.__future = aio.Future()
        self.__root = root

    def __getattr__(self, attr):
        return type(self)(self.__element, [*self.__sequence, attr], self.__root)

    __getitem__ = __getattr__

    def __call__(self, *args):
        return type(self)(self.__element, [*self.__sequence, args], self.__root)

    def __await__(self):
        self.__element.do(
            _extractor(self.__root, self.__sequence),
            future=self.__future,
        )
        return iter(self.__future)



import asyncio as aio
import base64
import inspect
import json
import os
import traceback
from contextvars import ContextVar
from functools import cached_property, wraps
from itertools import count, islice
from pathlib import Path
from uuid import uuid4 as uuid

from hrepr import H, Tag
from starlette.exceptions import HTTPException
from starlette.responses import (
    FileResponse,
    HTMLResponse,
    JSONResponse,
    PlainTextResponse,
    RedirectResponse,
    Response,
)
from starlette.routing import Mount, Route, WebSocketRoute
from starlette.websockets import WebSocket, WebSocketDisconnect

from .constructors import NamespaceDict, construct
from .page import Page
from .repr import Representer
from .templating import Template, template
from .utils import Queue, format_error, keyword_decorator, logger

here = Path(__file__).parent

_count = count()

dev_injections = []

debug_mode = ContextVar("debug_mode", default=int(os.environ.get("STARBEAR_DEBUG", 0)))

_gc_message = (
    "It may have been garbage-collected."
    " References in the HTML trees you create are weak,"
    " so you have to keep strong references to ensure they"
    " are not collected. Likely culprits are lambda"
    " expressions or nested functions."
    "\n\nAlternatively, you can pass strongrefs=True to"
    " @bear or the constructor to force references to be"
    " kept alive, if you are confident this will not leak"
    " memory."
)


bearlib_template = Template(here / "bearlib-template.html")


def routeinfo(params="", path=None, root=False, cls=Route, **kw):
    def deco(method):
        assert method.__name__.startswith("route_")
        name = method.__name__.removeprefix("route_")
        method.routeinfo = {
            "cls": cls,
            "name": name,
            "path": path or ("/" if root else f"/{name}"),
            "root": root,
            "params": params,
            "keywords": kw,
        }
        return method

    return deco


def gather_routes(obj):
    for method_name in dir(obj):
        if method_name.startswith("route_"):
            method = getattr(obj, method_name)
            routeinfo = method.routeinfo
            yield method, routeinfo


def autoroutes(defns, prefix, mangle, wrap=None):
    routes = []
    for method, routeinfo in defns:
        wmeth = wrap(method, routeinfo) if wrap else method
        mname = mangle(routeinfo["name"])
        route = routeinfo["cls"](
            prefix + routeinfo["path"] + routeinfo["params"],
            wmeth,
            name=mname,
            **routeinfo["keywords"],
        )
        routes.append(route)

        if routeinfo["root"]:
            route = routeinfo["cls"](
                "/",
                wmeth,
                name=mangle("root"),
                **routeinfo["keywords"],
            )
            routes.append(route)

    return routes


class AbstractBear:
    def __init__(self):
        self.appid = next(_count)
        self.route = None
        self.router = None
        self.representer = None
        self._json_decoder = json.JSONDecoder(object_pairs_hook=self.object_pairs_hook)

    ###########
    # Methods #
    ###########

    def mangle(self, name):
        return f"app{self.appid}_{name}"

    def path_for(self, name, **kwargs):
        return self.router.url_path_for(self.mangle(name), **kwargs)

    def object_pairs_hook(self, pairs):
        return NamespaceDict(pairs)

    async def json(self, request):
        body = await request.body()
        if isinstance(body, bytes):
            body = body.decode(encoding="utf8")
        return self._json_decoder.decode(body)

    def ensure_router(self, request):
        router = request.scope["router"]
        if self.router is None:
            self.router = router
            self.app = request.scope.get("app", None)
        else:
            assert self.router is router

    ##############################
    # For standalone application #
    ##############################

    def routes(self):
        return []

    @cached_property
    def _mnt(self):
        return Mount("/", routes=self.routes())

    async def __call__(self, scope, receive, send):
        await self._mnt.handle(scope, receive, send)


class BasicBear(AbstractBear):
    def __init__(self, template, template_params):
        super().__init__()
        self.route = None
        self._template = template
        self._template_params = {
            "title": "Starbear",
            "body": "",
            "connect_line": "",
            "bearlib": bearlib_template,
            **template_params,
        }

    ###################
    # Basic functions #
    ###################

    def template_asset(self, name, where):
        return self.route + "/file/" + self.representer.file_registry.register(where / name)

    def template(self, template_path=None, **params):
        template_path = template_path or self._template
        location = template_path.parent
        agg_params = {
            **self._template_params,
            "route": self.route,
            "dev": dev_injections,
            **params,
        }
        return template(
            template_path,
            **agg_params,
            _asset=lambda name: self.template_asset(name, location),
            _std=lambda name: self.template_asset(name, here),
        )

    def error_response(self, code, message, debug=None, exception=None):
        msg = format_error(
            message=message,
            debug=debug,
            exception=exception,
            show_debug=debug_mode.get(),
        )
        return JSONResponse({"message": msg}, status_code=code)

    ################
    # Basic routes #
    ################

    @routeinfo("/{method:int}", methods=["GET", "POST"])
    async def route_method(self, request):
        method_id = request.path_params["method"]
        try:
            method = self.representer.object_registry.resolve(method_id)
        except KeyError:
            return self.error_response(
                code=404,
                message="Application error: method not found.",
                debug=_gc_message,
            )
        try:
            args = await self.json(request)
        except json.JSONDecodeError:
            args = [await request.body()]
        try:
            result = method(*args, **request.query_params)
        except Exception as exc:
            return self.error_response(
                code=500,
                message="Application error.",
                exception=exc,
            )
        if inspect.iscoroutine(result):
            result = await result
        if isinstance(result, Tag):
            return HTMLResponse(self.representer(result))
        else:
            return JSONResponse(result)

    @routeinfo(methods=["POST"])
    async def route_post(self, request):
        data = await self.json(request)
        if "error" in data:
            self.representer.future_registry.reject(
                fid=data["reqid"],
                error=data["error"],
            )
        else:
            self.representer.future_registry.resolve(
                fid=data["reqid"],
                value=data.get("value", None),
            )
        return JSONResponse({"status": "ok"})

    @routeinfo(methods=["POST"])
    async def route_queue(self, request):
        data = await self.json(request)
        self.representer.queue_registry.put(
            qid=data["reqid"],
            value=data["value"],
        )
        return JSONResponse({"status": "ok"})

    @routeinfo("/{path:path}")
    async def route_file(self, request):
        pth = self.representer.file_registry.get_file_from_url(request.path_params["path"])
        if pth is None:
            raise HTTPException(status_code=404, detail="File not found or not available.")
        return FileResponse(pth, headers={"Cache-Control": "no-cache"})

    @routeinfo("/{path:path}")
    async def route_vfile(self, request):
        vf = self.representer.vfile_registry.get(request.path_params["path"])
        return Response(content=vf.content, media_type=vf.type)


class LoneBear(BasicBear):
    def __init__(self, fn, template=None, template_params={}, strongrefs=False):
        super().__init__(
            template=template or (here / "page-template.html"),
            template_params=template_params,
        )
        self.strongrefs = strongrefs
        self.fn = fn
        self.__doc__ = getattr(fn, "__doc__", None)

    ####################
    # Route generation #
    ####################

    def ensure_representer(self, request):
        self.ensure_router(request)
        if self.representer is None:
            self.route = self.path_for("main").rstrip("/")
            self.representer = Representer(self.route, strongrefs=self.strongrefs)

    def wrap_route(self, method, routeinfo):
        @wraps(method)
        async def wrapped(request):
            self.ensure_representer(request)
            return await method(request)

        return wrapped

    def routes(self):
        return autoroutes(
            defns=gather_routes(self),
            prefix="/!",
            mangle=self.mangle,
            wrap=self.wrap_route,
        )

    @routeinfo(root=True)
    async def route_main(self, request):
        response = await self.fn(request)
        if isinstance(response, Tag):
            if response.name != "html":
                response = self.template(body=response)
            html = self.representer.generate_string(response)
            return HTMLResponse(f"<!DOCTYPE html>\n{html}")
        elif isinstance(response, dict):
            return JSONResponse(response)
        else:
            return PlainTextResponse(str(response))


class Cub(BasicBear):
    def __init__(
        self,
        mother,
        process,
        query_params={},
        session={},
        template=None,
        template_params={},
        strongrefs=False,
    ):
        super().__init__(
            template=template or (here / "page-template.html"),
            template_params={"connect_line": "bear.connect()", **template_params},
        )
        self.mother = mother
        self.fn = mother.fn
        self.process = process
        self.query_params = query_params
        self.session = session
        self.route = self.mother.path_for("main", process=self.process).rstrip("/")
        self.representer = Representer(self.route, strongrefs=strongrefs)
        self.iq = Queue()
        self.oq = Queue()
        self.history = []
        self.reset = False
        self.ws = None
        self.page = Page(instance=self, debug=debug_mode.get())
        self.coro = aio.create_task(self.run())
        self.log("info", "Created process")

    def log(self, level, msg, **extra):
        try:
            user = self.session.get("user", {}).get("email", None)
        except:  # noqa: E722
            logger.error("Could not get user")
        getattr(logger, level)(msg, extra={"proc": self.process, "user": user, **extra})

    def destroy(self):
        self.coro.cancel()

    async def run(self):
        reason = "done"
        try:
            await self.fn(self.page)
            await self.page.sync()
        except aio.CancelledError:
            reason = "cancelled"
        except Exception as exc:
            reason = "error"
            self.log("error", str(exc), traceback=traceback)
            self.page.error(
                message=H.b("An error occurred. You may need to refresh the page."),
                exception=exc,
            )
        finally:
            self.log("info", f"Finished process ({reason})")

    def object_pairs_hook(self, pairs):
        dct = NamespaceDict(pairs)
        if "%" in dct:
            return construct(self.page, dct)
        else:
            return dct

    async def json(self, request):
        try:
            return await super().json(request)
        except KeyError as exc:
            self.page.error(
                message=f"Error constructing: reference {exc.args[0]} not found",
                debug=_gc_message,
            )
            raise
        except Exception as exc:
            self.page.error(
                message="Error constructing object",
                exception=exc,
            )
            raise

    ##############
    # Cub routes #
    ##############

    @routeinfo(root=True)
    async def route_main(self, request):
        self.mother.declare_active(self)
        node = self.template()
        self.reset = True
        html = self.representer.generate_string(node)
        return HTMLResponse(
            f"<!DOCTYPE html>\n{html}",
            headers={
                "Cache-Control": "no-cache, no-store, must-revalidate",
                "Pragma": "no-cache",
                "Expires": "0",
            },
        )

    @routeinfo(cls=WebSocketRoute)
    async def route_socket(self, ws):
        self.mother.declare_active(self)

        async def recv():
            while True:
                try:
                    data = await ws.receive_json()
                    self.iq.put_nowait(data)
                except WebSocketDisconnect:
                    break

        async def send():
            while True:
                obj, in_history = await self.oq.get()
                try:
                    await ws.send_json(obj)
                    if in_history:
                        self.history.append(obj)
                except RuntimeError:
                    # Put the unsent element back into the queue
                    self.oq.putleft((obj, in_history))
                    break

        if self.ws:
            try:
                await self.ws.close()
            except RuntimeError:
                pass

        await ws.accept()
        self.ws = ws

        if self.reset:
            for entry in self.history:
                await ws.send_text(entry)
            self.reset = False

        await aio.wait(
            [aio.create_task(recv()), aio.create_task(send())],
            return_when=aio.FIRST_COMPLETED,
        )
        self.mother.declare_dormant(self)


def get_process_from_request(request):
    process_base = request.path_params.get("process", None)
    if process_base is None:
        process_base = base64.urlsafe_b64encode(uuid().bytes).decode("utf8").strip("=")
    return process_base


class MotherBear(AbstractBear):
    def __init__(
        self,
        fn,
        soft_process_cap=1_000,
        hard_process_cap=1_000_000,
        hide_processes=True,
        **cub_params,
    ):
        super().__init__()
        self.fn = fn
        self.__doc__ = getattr(fn, "__doc__", None)
        self.router = None
        self.hard_process_cap = hard_process_cap
        self.soft_process_cap = soft_process_cap
        self.hide_processes = hide_processes
        self.cub_params = cub_params
        self.cubs = {}
        self.dormant_cubs = {}

    #############
    # Utilities #
    #############

    def _create_new_cub(self, proc, query_params, session):
        reclaims = max(0, min(len(self.dormant_cubs), len(self.cubs) - self.soft_process_cap))
        processes = list(islice(n=reclaims, iterable=self.dormant_cubs))

        for p in processes:
            cub = self.dormant_cubs.pop(p)
            cub.destroy()
            del self.cubs[p]

        if len(self.cubs) > self.hard_process_cap:
            raise Exception("Cannot serve request: too many processes exist.")

        self.cubs[proc] = Cub(
            self,
            proc,
            query_params=query_params,
            session=session,
            **self.cub_params,
        )

    def _get(self, proc, query_params={}, session={}, ensure=False):
        if proc not in self.cubs:
            if ensure:
                self._create_new_cub(
                    proc,
                    query_params=query_params,
                    session=session,
                )
            else:
                return None
        return self.cubs[proc]

    def declare_active(self, cub):
        self.dormant_cubs.pop(cub.process, None)

    def declare_dormant(self, cub):
        self.dormant_cubs[cub.process] = cub

    #################
    # Mother routes #
    #################

    @routeinfo(root=True)
    async def route_dispatch(self, request):
        self.ensure_router(request)
        process = get_process_from_request(request)
        main_path = self.path_for("main", process=process)
        if self.hide_processes:
            try:
                session = request.session
            except AssertionError:
                session = {}
            return await self._get(
                process, query_params=request.query_params, session=session, ensure=True
            ).route_main(request)
        else:
            url = main_path
            if request.query_params:
                url = f"{url}?{request.query_params}"
            return RedirectResponse(url=url)

    ####################
    # Construct routes #
    ####################

    def wrap_route(self, method, routeinfo):
        ensure = routeinfo["root"]

        @wraps(method)
        async def forward(request):
            self.ensure_router(request)
            process = get_process_from_request(request)
            cub = self._get(process, query_params=request.query_params, ensure=ensure)
            if cub is None:
                logger.warning(f"Trying to access missing process: {process}")
                if isinstance(request, WebSocket):
                    await request.accept()
                    await request.close(code=3002, reason="Missing application")
                else:
                    return JSONResponse(
                        {
                            "missing": process,
                            "message": "Session killed. Please refresh.",
                        },
                        status_code=404,
                    )
            else:
                return await method(cub, request)

        return forward

    def routes(self):
        return [
            *autoroutes(
                defns=gather_routes(self),
                prefix="",
                mangle=self.mangle,
            ),
            Mount(
                "/!{process:str}/",
                routes=autoroutes(
                    defns=gather_routes(Cub),
                    prefix="",
                    mangle=self.mangle,
                    wrap=self.wrap_route,
                ),
            ),
        ]


class ConfigurableBear(MotherBear):
    def __init__(self, config, **params):
        super().__init__(self.app, **params)
        self.config = config


class ConfigurableSimpleBear(LoneBear):
    def __init__(self, config, **params):
        super().__init__(self.app, **params)
        self.config = config


@keyword_decorator
def bear(fn, **kwargs):
    return MotherBear(fn, **kwargs)


def simplebear(fn, **kwargs):
    return LoneBear(fn, **kwargs)
